<!doctype html>
<html lang="en">

  <head>
    <title>The Movie Graph</title>
    <link rel="stylesheet" href="/static/main.css">
  </head>

  <body>

    <div class="header">
      <nav><strong>The Movie Graph</strong></nav>
    </div>

    <h1>The Movie Graph</h1>
    <ul>
        <li><a href="/person/">People</a></li>
        <li><a href="/movie/">Movies</a></li>
    </ul>

    <div class="footer">
      <code>(graphs)-[:ARE]->(everywhere)</code>
    </div>

  </body>

</html>

